/* partida.c
 * Autors: Sergio Blanco Cuaresma
 *         Ramn Quionero Bonmat
 *         Valentn Rey Gmez
 * Data: Gener del 2002
 * Descripci: Genera 2 fills, cadascun carga un proces vaixells per jugar entre
 * ells, un comena la partida i l'altre espera tir. A ms, genera un tercer fill
 * que carrega el proces xivato al qual sobornaran els 2 altres fills per que li
 * digui posicions dels vaixells del contrari. Comunicaci amb cues de missatges,
 * memria compartida, semfors i senyals.
 *
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <assert.h>
#include <signal.h>
#include "semafor.h"
#include "memoria.h"
#include "missatge.h"
#include "tipus.h"    // Per sizeof(FLOTA)
pid_t pid_f1, pid_f2, pid_f3, pid_term;
int pasta1, pasta2, sem_pasta1, sem_pasta2;
unsigned int *pasta1_ptr, *pasta2_ptr;
void (*oldsig2)(int i);
void (*oldsig14)(int i);

// Finalitzem tots el procesos fills, aixi el pare podr acabar correctament
void acabar(void){
  kill(pid_f1, 15);
  kill(pid_f2, 15);
  kill(pid_f3, 15);
}

// Incrementa els diners dels fills
void paga(void){

  waitS(sem_pasta1);
  if (*pasta1_ptr < 65535)
    *pasta1_ptr = *pasta1_ptr + 5000;
  signalS(sem_pasta1);

  waitS(sem_pasta2);
  if (*pasta2_ptr < 65535)
    *pasta2_ptr = *pasta2_ptr + 5000;
  signalS(sem_pasta2);

  signal(SIGALRM, (void *)paga);
  alarm(1);   // D'aqui a un segon rebre un altre cop la senyal SIGALRM
}


//Principal
int main (void) {
  bool sortir;
  int n, bucle;
  // Memoria compartida i semafors
  char pasta1_s[20], pasta2_s[20], sem_pasta1_s[20], sem_pasta2_s[20];
  int vaixells1, vaixells2, sem_vaixells1, sem_vaixells2;
  char vaixells1_s[20], vaixells2_s[20], sem_vaixells1_s[20], sem_vaixells2_s[20];
  // Missatges
  int fill1_a_2, fill2_a_1, fill1_a_xiv, fill2_a_xiv;
  char fill1_a_2_s[20], fill2_a_1_s[20], fill1_a_xiv_s[20], fill2_a_xiv_s[20];
  int xiv_a_fill1, xiv_a_fill2;
  char xiv_a_fill1_s[20], xiv_a_fill2_s[20];

  // Inicialitzem semafors i memria compartida
  // Semafor + Mem. compartida dels diners del fill 1
  sem_pasta1 = ini_sem(1);
  sprintf(sem_pasta1_s, "%i", sem_pasta1);
  pasta1 = ini_mem(sizeof(unsigned int));
  sprintf(pasta1_s, "%i", pasta1);            // converteixo el int. a string per poder-lo
  pasta1_ptr = map_mem(pasta1);
  *pasta1_ptr = 5000;
  
  // Semafor + Mem. compartida dels diners del fill 2
  sem_pasta2 = ini_sem(1);
  sprintf(sem_pasta2_s, "%i", sem_pasta2);
  pasta2 = ini_mem(sizeof(unsigned int));
  sprintf(pasta2_s, "%i", pasta2);
  pasta2_ptr = map_mem (pasta2);
  *pasta2_ptr = 5000;
 
  // Semafor + Mem. compartida dels VAIXELLS del fill 1
  sem_vaixells1 = ini_sem(1);
  sprintf(sem_vaixells1_s, "%i", sem_vaixells1);
  vaixells1 = ini_mem(sizeof(FLOTA));
  sprintf(vaixells1_s, "%i", vaixells1);            // converteixo el int. a string per poder-lo

  // Semafor + Mem. compartida dels VAIXELLS del fill 2
  sem_vaixells2 = ini_sem(1);
  sprintf(sem_vaixells2_s, "%i", sem_vaixells2);
  vaixells2 = ini_mem(sizeof(FLOTA));
  sprintf(vaixells2_s, "%i", vaixells2);
 
  // Cues de missatges entre fills
  fill1_a_2 = ini_mis();
  sprintf(fill1_a_2_s, "%i", fill1_a_2);
  fill2_a_1 = ini_mis();
  sprintf(fill2_a_1_s, "%i", fill2_a_1);
  
  // Cues de missatges entre xivato i fill1
  fill1_a_xiv = ini_mis();
  sprintf(fill1_a_xiv_s, "%i", fill1_a_xiv);
  xiv_a_fill1 = ini_mis();
  sprintf(xiv_a_fill1_s, "%i", xiv_a_fill1);

  // Cues de missatges entre xivato i fill2
  fill2_a_xiv = ini_mis();
  sprintf(fill2_a_xiv_s, "%i", fill2_a_xiv);
  xiv_a_fill2 = ini_mis();
  sprintf(xiv_a_fill2_s, "%i", xiv_a_fill2);

 
  // Cada segon s'executara la funcio paga();
  oldsig14 = (void *)signal(SIGALRM, (void *)paga);
  alarm(1);
  // Si provoquem la senyal SIGINT amb CTRL+C sortim correctament
  oldsig2 = (void *)signal(SIGINT, (void *)acabar);
  
  if ((pid_f1=fork())== (pid_t) 0)                // proces primer fill
  {
    // Cargem proces amb nom Gates que esperara un tir, mostrant missatges per pantalla i
    // que te els seus vaixells guardats a pos1.txt (tb li pasem tots els IPCs)
    execlp("./vaixells","./vaixells","Gates","0","0","pos1.txt", fill2_a_1_s, fill1_a_2_s, vaixells1_s, sem_vaixells1_s, pasta1_s, sem_pasta1_s, xiv_a_fill1_s, fill1_a_xiv_s, (char *)0);
    n=write(2,"ERROR en EXEC1\n",14);   // cas d'exec erroni
    assert(n==14);
    _exit(3);
  }
  else
  {
    if (pid_f1 > (pid_t) 0)                       // proces pare
    {
      if ((pid_f2=fork())== (pid_t) 0)                /* proces segon fill */
      {
        // Cargem proces amb nom Torvalds que comena partida, mostrant missatges per pantalla i
        // que te els seus vaixells guardats a posicions2.txt (tb li pasem tots els IPCs)
        execlp("./vaixells","./vaixells","Torvalds","1","0", "pos2.txt", fill1_a_2_s, fill2_a_1_s, vaixells2_s, sem_vaixells2_s, pasta2_s, sem_pasta2_s, xiv_a_fill2_s, fill2_a_xiv_s, (char *)0);
        n=write(2,"ERROR en EXEC2\n",14);
        assert(n==14);
        _exit(4);
      }
      else
      {
        if (pid_f2 > (pid_t) 0)                       // proces pare
        {

          if ((pid_f3=fork())== (pid_t) 0)        // Tercer fill (xivato)
          {
            // Cargem proces xivato indicant-li que els noms del jugadors son Gates i Torvalds (a ms de tots els IPCs)
            execlp("./xivato", "./xivato", fill1_a_xiv_s, xiv_a_fill1_s, fill2_a_xiv_s, xiv_a_fill2_s, vaixells1_s, sem_vaixells1_s, vaixells2_s, sem_vaixells2_s, pasta1_s, sem_pasta1_s, pasta2_s, sem_pasta2_s, "Gates", "Torvalds", (char *)0);
            n=write(2,"ERROR en EXEC2\n",14);
            assert(n==14);
            _exit(4);
          } else {
            if (pid_f3 > (pid_t) 0){                       // proces pare, ha d'esperar que acabin els fills
              // El proces pare espera a que els processos fills acabin
              for (bucle=0; bucle<3; bucle++){
                pid_term = wait(NULL);
                // Comprovem que realment ha sigut un fill el que ha acabat i no hem passat
                // degut a una senyal (com la SIGALRM) que fa que es saltin els waits a sistemes
                // amb solaris. Aquest codi tb es correcte per altres UNIX.
                while ((pid_term!=pid_f1) && (pid_term!=pid_f2) && (pid_term!=pid_f3)){
                  pid_term = wait(NULL);
                }
              }
              // Alliberem recursos (IPCs)
              elim_mem(pasta1);
              elim_sem(sem_pasta1);
              elim_mem(pasta2);
              elim_sem(sem_pasta2);
              elim_mem(vaixells1);
              elim_sem(sem_vaixells1);
              elim_mem(vaixells2);
              elim_sem(sem_vaixells2);
              elim_mis(fill1_a_xiv);
              elim_mis(fill2_a_xiv);
              elim_mis(xiv_a_fill1);
              elim_mis(xiv_a_fill2);
              elim_mis(fill1_a_2);
              elim_mis(fill2_a_1);
              signal(SIGINT, oldsig2);  // Recuperem l'antiga funci assiganda
              signal(SIGALRM, oldsig14);

              _exit(0);
            } else {
              n=write(2,"ERROR en FORK del tercer fill\n",14);
              assert(n==14);
              _exit(7);
            }
          }
        }
        else                               /* error en el fork del segon fill */
        {
          n=write(2,"ERROR en FORK del segon fill\n",14);
          assert(n==14);
          _exit(6);
        }
      }
    }
    else                               /* error en el fork del primer fill */
    {
     n=write(2,"ERROR en FORK del primer fill\n",14);
     assert(n==14);
     _exit(5);
    }
  }
}
  
  
