/* vaixells.c
 * Autors: Sergio Blanco Cuaresma
 *         Ramn Quionero Bonmat
 *         Valentn Rey Gmez
 * Data: Gener del 2002
 * Descripci: Execuci: ./vaixells nom_jugador 1 1 fitxer.txt IPCs... 
 *    NO S'HA D'EXECUTAR DIRECTAMENT, SINO A TRAVES DE PARTIDA!!
 *
 * A on el primer 1 significa que comena a jugar el programa (0 comena el jugador)
 * el segon 1 significa que mostri missatges per la sortida d'error (0 que no)
 * i fitxer.txt es el nom d'un fitxer que cont les posicions dels vaixells del programa.
 *
 * El programa juga al joc dels vaixells contra un altre proces comunicant-se per cues de missatges
 * en una taula de 10x10, fa tirs del estil A1, B5... espera rebre A, T o E segons si ha sigut Aigua,
 * Tocat o enfonsat, i tb llegeix els tirs del contrari retornant tb A, T o E. Si algun dels 2 ha
 * tocat o enfonsat un vaixell, continua disparant.
 *
 * Els tirs realitzats pel programa son els proposats per un tercer proces, anomenat xivato, al qual
 * s'intentar sobornar per que ens digui una posici correcta o per que enganyi al contrari.
 *
 */

#include <stdio.h>
#include <fcntl.h>
#include <assert.h>
#include <errno.h>
#include "tipus.h"
#include <stdlib.h>
#include <time.h>
#include "semafor.h"
#include "memoria.h"
#include "missatge.h"


#define N 10

#define BUIT 'B'
#define VAIXELL 'V'

#define AIGUA 'A'
#define TOCAT 'T'
#define ENFONSAT 'E'

#define POSLL(n) (n-'A')
#define POSN(n) (n-'0')
//#define POSN(n) (n)
#define LLPOS(n) (n+'A')
#define NPOS(n) (n+'0')

FLOTA *flota;
int sem_flota;
bool tirs[10][10];

// Llegeix un transatlantic
void llegir_transatlantic (FILE * posicions)
{
  char bucle;

  waitS(sem_flota);
  for (bucle = 0; bucle < 4; bucle++) {
    flota->transatlantic.pos[bucle].lletra = toupper (fgetc (posicions));
    flota->transatlantic.pos[bucle].numero = POSN (fgetc (posicions));
    flota->transatlantic.pos[bucle].tocat = FALSE;
  }
  flota->transatlantic.enfonsat = FALSE;       // Inicialitzem
  fgetc (posicions);            // End of line
  signalS(sem_flota);
}

// LLegueix dos buques
void llegir_buques (FILE * posicions)
{
  char bucle1, bucle2;

  waitS(sem_flota);
  for (bucle1 = 0; bucle1 < 2; bucle1++)        // Per cada buque
  {
    for (bucle2 = 0; bucle2 < 3; bucle2++)      // Per cada posici
    {
      flota->buque[bucle1].pos[bucle2].lletra = toupper(fgetc (posicions));
      flota->buque[bucle1].pos[bucle2].numero = POSN (fgetc (posicions));
      flota->buque[bucle1].pos[bucle2].tocat = FALSE;
    }
    flota->buque[bucle1].enfonsat = FALSE;     // Inicialitzem
    fgetc (posicions);          // End of line
  }
  signalS(sem_flota);
}

// Llegueix tres yots
void llegir_yots (FILE * posicions)
{
  char bucle1, bucle2;

  waitS(sem_flota);
  for (bucle1 = 0; bucle1 < 3; bucle1++)        // Per cada yot
  {
    for (bucle2 = 0; bucle2 < 2; bucle2++)      // Per cada posici
    {
      flota->yot[bucle1].pos[bucle2].lletra = toupper(fgetc (posicions));
      flota->yot[bucle1].pos[bucle2].numero = POSN (fgetc (posicions));
      flota->yot[bucle1].pos[bucle2].tocat = FALSE;
    }
    flota->yot[bucle1].enfonsat = FALSE;       // Inicialitzem
    fgetc (posicions);          // End of line
  }
  signalS(sem_flota);
}

// Llegueix quatre barques
void llegir_barques (FILE * posicions)
{
  char bucle1, bucle2;

  waitS(sem_flota);
  for (bucle1 = 0; bucle1 < 4; bucle1++)        // Per cada barca
  {
    flota->barca[bucle1].pos.lletra = toupper(fgetc (posicions));
    flota->barca[bucle1].pos.numero = POSN (fgetc (posicions));
    flota->barca[bucle1].pos.tocat = FALSE;
    flota->barca[bucle1].enfonsat = FALSE;
    getc (posicions);           // End of line
  }
  signalS(sem_flota);
}


// Llegueix totes les posicions del vaixells
void llegir_fitxer (char *fitxer)
{
  FILE *posicions;

  if ((posicions = fopen (fitxer, "r")) == NULL) {
    perror ("posicions.txt");
    exit (1);
  }

  llegir_transatlantic (posicions);
  llegir_buques (posicions);
  llegir_yots (posicions);
  llegir_barques (posicions);

  fclose (posicions);
}

// Comprova si ens han tocat, enfonsat o fet aigua a partir d'una posici indicada com argument
char comprovar_tir (char *resposta_tir)
{
  char bucle, bucle2;

  waitS(sem_flota);
  // Si ja esta enfonsat no cal comprovar si ha sigut tocat
  if (flota->transatlantic.enfonsat == FALSE) {
    bucle = 0;
    // Mentre no trobem cap posici igual, avancem
    while (((flota->transatlantic.pos[bucle].numero != POSN (resposta_tir[1]))
            || (flota->transatlantic.pos[bucle].lletra != resposta_tir[0]))
           && (bucle < 3)) {
      bucle++;
    }
    // Si hem sortit pq hi ha una posici igual
    if ((flota->transatlantic.pos[bucle].numero == POSN (resposta_tir[1]))
        && (flota->transatlantic.pos[bucle].lletra == resposta_tir[0])) {
      flota->transatlantic.pos[bucle].tocat = TRUE;    // Ens han tocat
      // Si totes les posicions estan tocades
      if ((flota->transatlantic.pos[0].tocat == TRUE) && (flota->transatlantic.pos[1].tocat == TRUE) && (flota->transatlantic.pos[2].tocat == TRUE) && (flota->transatlantic.pos[3].tocat == TRUE)) {
        flota->transatlantic.enfonsat = TRUE;  // Ens han enfonsat
        signalS(sem_flota);
        return ENFONSAT;
      }
      signalS(sem_flota);
      return TOCAT;
    }
  }
  // Per cada buque (2 en total)
  for (bucle2 = 0; bucle2 < 2; bucle2++) {
    // Si ja esta enfonsat no cal comprovar si ha sigut tocat
    if (flota->buque[bucle2].enfonsat == FALSE) {
      bucle = 0;
      // Mentre no trobem cap posici igual, avancem
      while (((flota->buque[bucle2].pos[bucle].numero != POSN (resposta_tir[1]))
              || (flota->buque[bucle2].pos[bucle].lletra != resposta_tir[0]))
             && (bucle < 2)) {
        bucle++;
      }
      // Si hem sortit pq hi ha una posici igual
      if ((flota->buque[bucle2].pos[bucle].numero == POSN (resposta_tir[1]))
          && (flota->buque[bucle2].pos[bucle].lletra == resposta_tir[0])) {
        flota->buque[bucle2].pos[bucle].tocat = TRUE;  // Ens han toca
        // Si totes les posicions estan tocades
        if ((flota->buque[bucle2].pos[0].tocat == TRUE) && (flota->buque[bucle2].pos[1].tocat == TRUE) && (flota->buque[bucle2].pos[2].tocat == TRUE)) {
          flota->buque[bucle2].enfonsat = TRUE;        // Ens han enfonsat
          signalS(sem_flota);
          return ENFONSAT;
        }
        signalS(sem_flota);
        return TOCAT;
      }
    }
  }

  // Per cada yot (3 en total)
  for (bucle2 = 0; bucle2 < 3; bucle2++) {
    // Si ja esta enfonsat no cal comprovar si ha sigut tocat
    if (flota->yot[bucle2].enfonsat == FALSE) {

      // Comprovem si es tracta de la primera posici
      if ((flota->yot[bucle2].pos[0].numero == POSN (resposta_tir[1]))
          && (flota->yot[bucle2].pos[0].lletra == resposta_tir[0])) {
        flota->yot[bucle2].pos[0].tocat = TRUE;        // Ens han toca
        if ((flota->yot[bucle2].pos[0].tocat == TRUE) && (flota->yot[bucle2].pos[1].tocat == TRUE)) {
          flota->yot[bucle2].enfonsat = TRUE;  // Ens han enfonsat
          signalS(sem_flota);
          return ENFONSAT;
        }
        signalS(sem_flota);
        return TOCAT;
      }
      // Comprovem si es tracta de la segona posici
      if ((flota->yot[bucle2].pos[1].numero == POSN (resposta_tir[1]))
          && (flota->yot[bucle2].pos[1].lletra == resposta_tir[0])) {
        flota->yot[bucle2].pos[1].tocat = TRUE;        // Ens han toca
        if ((flota->yot[bucle2].pos[0].tocat == TRUE) && (flota->yot[bucle2].pos[1].tocat == TRUE)) {
          flota->yot[bucle2].enfonsat = TRUE;  // Ens han enfonsat
          signalS(sem_flota);
          return ENFONSAT;
        }
        signalS(sem_flota);
        return TOCAT;
      }

    }
  }

  // Per cada barca (4 en total)
  for (bucle2 = 0; bucle2 < 4; bucle2++) {
    // Si ja esta enfonsat no cal comprovar si ha sigut tocat// Si ja esta enfonsat no cal comprovar si ha sigut tocat
    if (flota->barca[bucle2].enfonsat == FALSE) {
      if ((flota->barca[bucle2].pos.numero == POSN (resposta_tir[1]))
          && (flota->barca[bucle2].pos.lletra == resposta_tir[0])) {
        flota->barca[bucle2].pos.tocat = TRUE; // Tocat
        flota->barca[bucle2].enfonsat = TRUE;  // i enfonsat
        signalS(sem_flota);
        return ENFONSAT;
      }
    }
  }
  // Si no ha retornat cap enfonsat o tocat en aquest punt
  // llavors es tracta d'un lloc a on no hi ha cap vaixell o
  // ja ha sigut tocat abans (situaci que no s'hauria de donar si
  // el programa juga amb si mateix degut al disseny)
  signalS(sem_flota);
  return AIGUA;
}


// Retorna true si tots els vaixells que tenim estan enfonsats, false en cas contrari
bool tots_enfonsats (void)
{
  waitS(sem_flota);
  if ((flota->transatlantic.enfonsat) && (flota->buque[0].enfonsat) && (flota->buque[1].enfonsat) && (flota->yot[0].enfonsat) && (flota->yot[1].enfonsat)
      && (flota->yot[2].enfonsat) && (flota->barca[0].enfonsat) && (flota->barca[1].enfonsat) && (flota->barca[2].enfonsat) && (flota->barca[3].enfonsat)) {
    signalS(sem_flota);
    return TRUE;
  } else {
    signalS(sem_flota);
    return FALSE;
  }
}


// Cos principal
int main (int argc, char *argv[])
{
  char nom_jugador[20], lletra, numero, resposta, resposta_tir[2], c;
  char begin, err, encerts = 0, direccio = 0;
  char i, j;
  char missatge[2], oferim[20];
  int sms_a_Jugador, sms_de_Jugador, sms_a_Xivato, sms_de_Xivato, sem_diners;
  unsigned int *diners, oferim_num;

  setbuf (stdin, NULL);         // anulem el buffer de l'entrada estandard
  setbuf (stdout, NULL);        // anulem el buffer de sortida estandard

  // Tractem arguments
  if (argc != 13) {
    fprintf (stderr, "Error: no es pot executar directament aquest arxiu\n");
    exit (1);
  }

  strncpy (nom_jugador, argv[1], 20);
  begin = atoi (argv[2]);
  err = atoi (argv[3]);
  if ((begin != 0) && (begin != 1)) {
    fprintf (stderr, "\n\tError, en els bools qui_comena i sortida_error\n");
  }
 
  // IPCs
  sms_de_Jugador = atoi(argv[5]);
  sms_a_Jugador = atoi (argv[6]);
  flota = map_mem (atoi (argv[7]));
  sem_flota = atoi (argv[8]);
  diners = map_mem (atoi (argv[9]));
  sem_diners = atoi(argv[10]);
  sms_de_Xivato = atoi(argv[11]);
  sms_a_Xivato = atoi(argv[12]);
  
  // Obtenim les posicions dels vaixells de la mquina
  llegir_fitxer (argv[4]);

  // Llavor
  srand (getpid ());            // Per la part 2-2 de la prctica

  // Mentre no hi hagui cap guanyador
  do {
    if (begin == 1)             // Torn de la mquina
    {
      oferim_num = abs ((char) (rand ()) % (*diners/3));  // Oferir com a mxim un ter del que tenim
      sprintf(oferim, "%u\0", oferim_num);
      // Sobornem al xivato
      sendM(sms_a_Xivato, oferim, strlen(oferim)+1);
      // Esperem tir que ens recomana fer
      receiveM(sms_de_Xivato, missatge);
      lletra = missatge[0];
      numero = POSN(missatge[1]);
      
      if (err == 1) {
        fprintf (stderr, "%s: Fem tir %c%i\n", nom_jugador, lletra, numero);
      }
      // Realitzem tir
      sendM (sms_a_Jugador, missatge, strlen(missatge)+1);
      // Esperem resposta
      receiveM (sms_de_Jugador, missatge);
      resposta = missatge[0];
      switch (resposta) {
        case TOCAT:
          break;
        case ENFONSAT:
          encerts++;            // Encerts en serveix per saber si hem enfonsat tots els seus vaixells
          break;                // El torn encara es nostre
        case AIGUA:
          begin = 0;            // Torn de l'altre
      }
    } else {                    // Torn de l'usuari
      if (err == 1) {
        fprintf (stderr, "%s: Esperem tir\n", nom_jugador);
      }
      oferim_num = abs ((char) (rand ()) % (*diners/3));  // Oferir com a mxim un ter del que tenim
      sprintf(oferim, "%u\0", oferim_num);
      // Sobornem al xivato per que enganyi a l'altre jugador
      sendM(sms_a_Xivato, oferim, strlen(oferim)+1);
      // Esperem tir
      receiveM (sms_de_Jugador, resposta_tir);
      resposta_tir[0] = toupper(resposta_tir[0]);

      // Comprovem si ens ha tocat, enfonsat o ha fet aigua
      c = comprovar_tir (resposta_tir);
      if (err == 1) {
        fprintf (stderr, "%s: Ens han donat a %c\n", nom_jugador, c);
      }

      sprintf(missatge, "%c\0", c);
      sendM (sms_a_Jugador, missatge, strlen(missatge)+1);
      if (c == AIGUA) {
        begin = 1;              // Torn de l'altre
      }
    }

  } while ((tots_enfonsats () == FALSE) && (encerts < 10));

  if (tots_enfonsats ()) {      // Perdem, ens han enfonsat tots els vaixells
    if (err == 1) {
      fprintf (stderr, "%s: Perd!\n", nom_jugador);
      fprintf(stderr, "Diners perdedor: %i\n", *diners);
    }
  } else {                      // Guanyem
    if (err == 1) {
      fprintf (stderr, "%s: Guanya!\n", nom_jugador);
      fprintf(stderr, "Diners guanyador: %i\n", *diners);
    }

  }

  return 0;
}
