/* xivato.c
 * Autors: Sergio Blanco Cuaresma
 *         Ramn Quionero Bonmat
 *         Valentn Rey Gmez
 * Data: Gener del 2002
 * Descripci: Agafa els diners que ofereixen 2 processos jugadors i proporciona
 * posicions del vaixells del contrincant de forma correcta o incorrecta segons la
 * quantitat de diners disposat a donar cada jugador.
 *
 * A ms visualitza per pantalla a cada canvi de torn, els tirs realitzats, el diners
 * i el nombre d'encerts de cada jugador.
 *
 */
#include <stdio.h>
#include <signal.h>
#include "tipus.h"
#include "semafor.h"
#include "missatge.h"
#include "memoria.h"

#define POSLL(n) (n-'A')
#define POSN(n) (n-'0')
#define LLPOS(n) (n+'A')
#define NPOS(n) (n+'0')

void (*oldsig)(int i);

// Busca a una flota determinada si la posici indicada hi ha algun vaixell (retorna TRUE)
bool hi_ha_vaixell (FLOTA *flota, int sem_flota, char *pos){
  char bucle, bucle2;

  waitS(sem_flota);
  // TRANSATLANTIC
    for (bucle=0; bucle<4; bucle++){
      if ((flota->transatlantic.pos[bucle].numero == pos[1])
          && (flota->transatlantic.pos[bucle].lletra == pos[0])) {
        signalS(sem_flota);
        return TRUE;
      }
    }
    
  // 2 BUQUES
  for (bucle2 = 0; bucle2 < 2; bucle2++) {
      for (bucle=0; bucle<3; bucle++){
        if ((flota->buque[bucle2].pos[bucle].numero == pos[1])
            && (flota->buque[bucle2].pos[bucle].lletra == pos[0])){
          signalS(sem_flota);
          return TRUE;
        }
      }
  }
  // 3 YOTS
  for (bucle2 = 0; bucle2 < 3; bucle2++) {
      for (bucle=0; bucle<2; bucle++){
        if ((flota->yot[bucle2].pos[bucle].numero == pos[1])
            && (flota->yot[bucle2].pos[bucle].lletra == pos[0])){
          signalS(sem_flota);
          return TRUE;
        }
      }
  }
  // 4 BARQUES
  for (bucle = 0; bucle < 4; bucle++) {
    if ((flota->barca[bucle].pos.numero == pos[1])
        && (flota->barca[bucle].pos.lletra == pos[0])){
      signalS(sem_flota);
      return TRUE;
    }
  }
  signalS(sem_flota);
  return FALSE;
}

  
// Troba una posici a on hi ha un vaixell per disparar (la guarda a *pos)
void trobar_pos(FLOTA *flota, int sem_flota, char *pos){
  char bucle, bucle2;

  pos[2] = '\0';
  waitS(sem_flota);
  // Si ja esta enfonsat no cal buscar posici per retornar
  // TRANSATLANTIC
  if (flota->transatlantic.enfonsat == FALSE) {
    bucle = 0;
    while ((flota->transatlantic.pos[bucle].tocat) && (bucle <3)){
      bucle++;
    }
    if (!flota->transatlantic.pos[bucle].tocat){
      pos[0] = flota->transatlantic.pos[bucle].lletra;
      pos[1] = flota->transatlantic.pos[bucle].numero;
      signalS(sem_flota);
      return;
    }
  }
  // 2 BUQUES
  for (bucle2 = 0; bucle2 < 2; bucle2++) {
    if (flota->buque[bucle2].enfonsat == FALSE) {
      bucle = 0;
      while ((flota->buque[bucle2].pos[bucle].tocat) && (bucle <2)){
        bucle++;
      }
      if (!flota->buque[bucle2].pos[bucle].tocat){
        pos[0] = flota->buque[bucle2].pos[bucle].lletra;
        pos[1] = flota->buque[bucle2].pos[bucle].numero;
        signalS(sem_flota);
        return;
      }
    }
  }
  // 3 YOTS
  for (bucle2 = 0; bucle2 < 3; bucle2++) {
    if (flota->yot[bucle2].enfonsat == FALSE) {
      bucle = 0;
      while ((flota->yot[bucle2].pos[bucle].tocat) && (bucle <3)){
        bucle++;
      }
      if (!flota->yot[bucle2].pos[bucle].tocat){
        pos[0] = flota->yot[bucle2].pos[bucle].lletra;
        pos[1] = flota->yot[bucle2].pos[bucle].numero;
        signalS(sem_flota);
        return;
      }
    }
  }
  // 4 BARQUES
  for (bucle = 0; bucle < 4; bucle++) {
    if ((!flota->barca[bucle].pos.tocat) && (bucle <4)){
      pos[0] = flota->barca[bucle].pos.lletra;
      pos[1] = flota->barca[bucle].pos.numero;
      signalS(sem_flota);
      return;
    }
  }
  signalS(sem_flota);
  return;
}

// Finalitzem el programa
void acabar(void){
  mvcur(0, COLS-1, LINES-1, 0);
  endwin();
  signal(SIGINT, oldsig);
}

// Principal
int main (int argc, char *argv[]){
  int sms_a_Jugador1, sms_de_Jugador1, sms_a_Jugador2, sms_de_Jugador2;
  FLOTA *flota1, *flota2;
  unsigned int *diners1, *diners2;
  int sem_flota1, sem_flota2, sem_diners1, sem_diners2;
  char message[5], lletra, numero, display[80], tirs1[80], tirs2[80];
  int oferta1, oferta2, torn=2, tocats1=0, tocats2=0;
  
  if (argc !=15) {
    fprintf(stderr, "Error: no es pot executar directament aquest arxiu %s\n", argv[0]);
    exit(1);
  }

  // Inicialitzem la pantalla (curses)
  if (initscr() == NULL) {
    fprintf(stderr, "Error a la creaci de l'entorn\n");
  }
  oldsig = (void *)signal(SIGINT, (void *)acabar);
  cbreak();
  noecho();
  nonl();
  intrflush(stdscr, FALSE);
  keypad(stdscr, TRUE);
  
  // IPCs
  sms_de_Jugador1 = atoi (argv[1]);
  sms_a_Jugador1 = atoi (argv[2]);
  sms_de_Jugador2 = atoi (argv[3]);
  sms_a_Jugador2 = atoi (argv[4]);
  flota1 = map_mem (atoi (argv[5]));
  sem_flota1 = atoi (argv[6]);
  flota2 = map_mem (atoi (argv[7]));
  sem_flota2 = atoi (argv[8]);
  diners1 = map_mem(atoi (argv[9]));
  sem_diners1 = atoi (argv[10]);
  diners2 = map_mem (atoi (argv[11]));
  sem_diners2 = atoi (argv[12]);

  // Llavor per generar aleatoris
  srand (getpid ());

  // Estetica
  mvprintw(0, 0, ">>>>>>>>>> [    ] %s", argv[13]);
  mvprintw(6, 0, ">>>>>>>>>> [torn] %s", argv[14]);
  sprintf(tirs1, "\0");  // Inicialitzem
  sprintf(tirs2, "\0");
  
  // Mentre no guanyi cap dels 2 processos
  while ((tocats1 <20) && (tocats2 <20)){
    // Esperem diners del primer jugador
    receiveM(sms_de_Jugador1, message);
    oferta1 = atoi(message);
    
    // Li treiem
    waitS(sem_diners1);
    *diners1 = *diners1 - oferta1;
    sprintf(display, "Diners: %i euros     \0", *diners1);
    signalS(sem_diners1);
    // Visualitzem diners
    mvaddstr(1, 0, display);
    // i nombre de blancs
    sprintf(display, "Ha fet blanc: %i vegades\0", tocats2);
    mvaddstr(2, 0, display);
    
    // Esperem diners del segon jugador
    receiveM(sms_de_Jugador2, message);
    oferta2 = atoi(message);

    // Li treiem
    waitS(sem_diners2);
    *diners2 = *diners2 - oferta2;
    sprintf(display, "Diners: %i euros     \0", *diners2);
    signalS(sem_diners2);
    // Visualitzem diners
    mvaddstr(7, 0, display);
    // i nombre de blancs
    sprintf(display, "Ha fet blanc: %i vegades\0", tocats1);
    mvaddstr(8, 0, display);

    
    if (torn == 1){   // Jugador 1 ens solicita posici
      if (oferta1 >= oferta2){    // Li donarem una posici a on hi ha un vaixell
        trobar_pos(flota2, sem_flota2, message);
        message[1] = NPOS(message[1]);
        sendM(sms_a_Jugador1, message, strlen(message)+1);
        tocats2++;
        // Guardem tir a la llista entre < > per saber que aquest tir es blanc segur
        sprintf(tirs1, "%s <%s>", tirs1, message);
      } else {    // Li donarem una posici incorrecte (ser aigua i el torn pasar a ser de l'altre)
        // Busquem una posici aleatoria a on no hi hagui vaixell, no importa si ja ha
        // disparat.
        message[0] = LLPOS (abs ((char) (rand ()) % 10));
        message[1] = abs ((char) (rand ()) % 10);
        while (hi_ha_vaixell(flota2, sem_flota2, message)){
          message[0] = LLPOS (abs ((char) (rand ()) % 10));
          message[1] = abs ((char) (rand ()) % 10);
        }
        message[1] = NPOS(message[1]);
        message[2] = '\0';
        sendM(sms_a_Jugador1, message, strlen(message)+1);
        // Canvi de torn, pq ha fet aigua
        torn = 2;

        // Visualitzem el canvi de torn i els tirs que ha fet
        mvaddstr(6, 12, "torn");
        mvaddstr(0, 12, "    ");
        mvaddstr(3, 0, "Tirs:                                                                          ");
        sprintf(tirs1, "%s -%s-", tirs1, message);
        mvaddstr(3, 6, tirs1);
        sprintf(tirs1, "\0");
        refresh();
        // Esperem que pulsi una tecla
        getch();
      }
    } else {    // Jugador 2 ens solicita posici
      if (oferta2 >= oferta1){    // Li donarem una posici a on hi ha un vaixell
        trobar_pos(flota1, sem_flota1, message);
        message[1] = NPOS(message[1]);
        sendM(sms_a_Jugador2, message, strlen(message)+1);
        tocats1++;
        // Guardem tir a la llista entre < > per saber que aquest tir es blanc segur
        sprintf(tirs2, "%s <%s>", tirs2, message);
      } else {    // Li donarem una posici incorrecte (ser aigua i el torn pasar a ser de l'altre)
        // Busquem una posici aleatoria a on no hi hagui vaixell, no importa si ja ha
        // disparat.
        message[0] = LLPOS (abs ((char) (rand ()) % 10));
        message[1] = abs ((char) (rand ()) % 10);
        while (hi_ha_vaixell(flota1, sem_flota1, message)){
          message[0] = LLPOS (abs ((char) (rand ()) % 10));
          message[1] = abs ((char) (rand ()) % 10);
        }
        message[1] = NPOS (message[1]);
        message[2] = '\0';
        sendM(sms_a_Jugador2, message, strlen(message)+1);
        // Canvi de torn, pq ha fet aigua
        torn = 1;

        // Visualitzem el canvi de torn i els tirs que ha fet
        mvaddstr(6, 12, "    ");
        mvaddstr(0, 12, "torn");
        mvaddstr(9, 0, "Tirs:                                                                          ");
        sprintf(tirs2, "%s -%s-", tirs2, message);
        mvaddstr(9, 6, tirs2);
        sprintf(tirs2, "\0");
        refresh();
        // Esperem que pulsi una tecla
        getch();
      }
    }
  }

  // Visualitzaci final, actualitzem tota la pantalla amb les dades finals de cada jugador
  if (tocats2 == 20) {  // Guanya Jugador 1
    mvaddstr(0, 11, "GuaNYa");
    mvaddstr(6, 11, " PeRD ");
    mvaddstr(3, 0, "Tirs:                                                                          ");
    mvaddstr(3, 6, tirs1);
  } else {    // Guanya Jugador 2
    mvaddstr(0, 11, " PeRD ");
    mvaddstr(6, 11, "GuaNYa");
    mvaddstr(9, 0, "Tirs:                                                                          ");
    mvaddstr(9, 6, tirs2);
  }

  // JUGADOR 1 //
  sprintf(display, "Ha fet blanc: %i vegades\0", tocats2);
  mvaddstr(2, 0, display);

  // Secci crtica
  waitS(sem_diners1);
  sprintf(display, "Diners: %i euros     \0", *diners1);
  signalS(sem_diners1);

  // Visualitzem diners
  mvaddstr(1, 0, display);
  
  // JUGADOR 2 //
  sprintf(display, "Ha fet blanc: %i vegades\0", tocats1);
  mvaddstr(8, 0, display);

  // Secci crtica
  waitS(sem_diners2);
  sprintf(display, "Diners: %i euros     \0", *diners2);
  signalS(sem_diners2);

  // Visualitzem diners
  mvaddstr(7, 0, display);

  refresh();
  getch();
  acabar();
  return 0;
}
